from __future__ import annotations
from anytree import NodeMixin
from datetime import datetime, timezone
from dotenv import load_dotenv
from os import environ
from os.path import join, dirname
from typing import Tuple, List, Any
import re
import requests
from rich.box import Box


__all__ = [
    "get_token", "get_url_info", "human_size", "humanize_time",
    "populate_tree", "ROUNDED_BORDER", "run_query", "set_token", "sort_entries"
]


ROUNDED_BORDER: Box = Box(
    """\
╭──╮
│  │
│  │
│  │
│  │
│  │
│  │
╰──╯
"""
)


def get_token():
    dotenv_path = join(dirname(__file__), '.env')
    load_dotenv(dotenv_path)

    return environ.get("GITSORT_TOKEN")


def set_token(token):
    """
    Set your Github personal access token in order to access
    private repositories and extend the usage of the GraphQL API.
    """
    import os
    from dotenv import load_dotenv
    from os.path import join, dirname

    dotenv_path = join(dirname(__file__), '.env')
    load_dotenv(dotenv_path)

    gitsort_token = os.environ.get("GITSORT_TOKEN")
    if not gitsort_token:
        with open(dotenv_path, "w") as f:
            f.write(f"GITSORT_TOKEN={token}")
        print("Github Token set!")
    else:
        inp = input("Github token already set! Do you want to update it? [y/n] ").lower()
        while inp not in ["y", "n"]:
            print("Invalid answer")
            inp = input("Github token already set! Do you want to update it? [y/n] ").lower()
        if inp == "y":
            with open(dotenv_path, "w") as f:
                f.write(f"GITSORT_TOKEN={token}")
            print("Github Token updated!")


def run_query(
    query: str,
    token: str,
    variables: dict | None = None,
    headers: dict | None = None
) -> Tuple[dict, str]:
    """
    Runs a Github GraphQL query and returns the result

    :param query: str
        GraphQL query
    :param token: str
        The users Github Personal Access Token
    :param variables: dict
        GraphQL Variables
    :param headers: dict
        Request headers
    :return: tuple
        The response and rate limit
    """
    if not headers:
        headers = {"Authorization": f"Bearer {token}"}

    request = requests.post(
        'https://api.github.com/graphql',
        json={'query': query, 'variables': variables},
        headers=headers
    )
    if request.status_code == 200:
        return request.json(), request.headers["X-RateLimit-Remaining"]
    else:
        raise Exception("Query failed to run by returning code of {}. {}".format(request.status_code, query))


def get_url_info(url: str) -> Tuple[str, str] | List[str]:
    """
    Retrieves owner and repository from a string
    :param url: str
        Either some form of Github Url or path such as `user/repo/whatever`
    :return: tuple | list
        Tuple containing owner and repo
    """
    is_link = re.compile(r"^(git(hub)?|https?)")
    is_git_path = re.compile(r"^[a-zA-Z0-9\-_.]+/[a-zA-Z0-9\-_.]+")
    git_url_regex = re.compile(r"^(https|git)?(://|@)?([^/:]+)[/:](?P<owner>[^/:]+)/(?P<name>.+)(.git)?$")
    is_git_repo = re.compile(r"((.git)|/)$")

    if is_link.match(url):
        if is_git_path.match(url):
            return url.split("/")[:2]

        match = git_url_regex.match(url)
        if not match:
            raise Exception("Invalid path")

        name = match.group("name").split("/")[0]
        name = is_git_repo.sub("", name)
        owner = match.group("owner")

        return owner, name
    else:
        if url.count("/") > 0:
            return url.split("/")[:2]
        raise Exception("Link/path must contain both user and repo")


def humanize_time(time_str):
    """
    Convert datetime into a more human-friendly format
    """
    if not time_str:
        return "null"

    now = datetime.now()
    date = datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%SZ")
    date = date.replace(tzinfo=timezone.utc)

    diff = int(now.timestamp() - date.timestamp())
    times = [
        1, 60, 3600, 86400, 604800, 2629746, 31556925
    ]
    times_str = [
        "Second", "Minute", "Hour", "Day", "Week", "Month", "Year"
    ]
    temp = [diff // t for t in times][::-1]
    for i, t in enumerate(temp):
        if t != 0:
            return f"{t} {times_str[6-i]}{'' if t == 1 else 's'} ago"


def human_size(bytes: int | float, units=None) -> str:
    """
    Convert bytes into a more human-friendly format
    """
    if units is None:
        units = ['bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB']
    return f"{round(bytes, 2)} " + units[0] if bytes < 1024 else human_size(bytes / 1024, units[1:])


class FileEntry(NodeMixin):
    def __init__(self, name: str, size: str = None, parent=None, children=None) -> None:
        super(FileEntry, self).__init__()

        if size != None:
            self.name = f"{name} ([green]{human_size(size)}[/])"
        else:
            self.name = f"[blue]{name}/[/]"

        self.parent = parent
        if children:
            self.children = children


def populate_tree(root_name: str, data: dict) -> "anytree.Node":
    """
    Populate the tree
    :param root_name: str
        Name of root node
    :param data: dict
        Data
    :return: anytree.node
    """
    root = FileEntry(root_name)

    def edges(tree, parent=None):
        for entry in tree:
            if entry["type"] == "blob":
                _ = FileEntry(entry["name"], entry["object"]["byteSize"], parent=parent)
            else:
                node = FileEntry(entry["name"], parent=parent)
                if entry["object"]:
                    edges(entry["object"]["entries"], parent=node)

    edges(data, root)
    return root


class Reversor:
    def __init__(self, obj: Any) -> None:
        self.obj = obj

    def __eq__(self, other: Any) -> bool:
        return other.obj == self.obj

    def __lt__(self, other: Any) -> bool:
        return other.obj < self.obj


def sort_entries(entries: List[Any]) -> List[Any]:
    """
    Recursively sort the data first based on type
    then alphabetically
    :param entries:
    :return:
    """
    entries = sorted(
        entries, key=lambda x: (
            Reversor(x["type"]),    # First sort by type (reversed)
            x["name"].lower()       # Then sort by alphabetical
        )
    )
    for entry in entries:
        if entry["type"] == "tree" and entry["object"]:
            entry["object"]["entries"] = sort_entries(entry["object"]["entries"])
    return entries



from .utils import *
from anytree import RenderTree, ContRoundStyle
from rich.table import Table
from rich import print
import click


INFO_QUERY = """\
query($owner: String!, $repo: String!, $branch: String!) {
    repository(owner: $owner, name: $repo) {
    createdAt
    diskUsage
    forkCount
    isArchived
    isDisabled
    isFork
    isInOrganization
    isLocked
    isMirror
    isPrivate
    issues {totalCount}
    languages(
        first: 1
        orderBy: {field:SIZE direction:DESC}
    ) {nodes { name }}
    latestRelease {
      name
      url
    }
    licenseInfo { url spdxId }
    object(expression:$branch) {
    ... on Commit {
            history {
                totalCount
            }
        }
    }
    openIssues:issues(states:OPEN) { totalCount }
    owner { login url }
    closed_pr: pullRequests(states:CLOSED) {totalCount}
    merged_pr: pullRequests(states:MERGED) {totalCount}
    open_pr: pullRequests(states:OPEN) {totalCount}
    pushedAt
    stargazerCount
    updatedAt
    url
    watchers {totalCount}
    }
}
"""


LANG_QUERY = """\
query($owner: String!, $repo: String!) {
    repository(owner: $owner, name: $repo) {
        languages(
            first: 100
            orderBy: {field:SIZE direction:DESC}
        ){
            totalSize
            totalCount
            edges {
                node {name color}
                size
            }
        }
    }
}
"""


FILE_QUERY_1 = """\
query RepoFiles($owner: String!, $repo: String!, $path: String!) {
repository(owner: $owner, name: $repo) {
    object(expression: $path) {
        ... on Tree {
            entries {
                name
                type
                object {
                    ... on Blob {
                        byteSize
                    }
"""


FILE_QUERY_2 = """\
                    }
                }
            }
        }
    }
}
"""


def Bool(input_bool: bool) -> str:
    if input_bool:
        return "[italic green]True[/]"
    else:
        return "[italic bright_red]False[/]"


def Number(input_num: int) -> str:
    return f"[magenta]{input_num}[/]"


def Date(input_date: str) -> str:
    return f"[#A9E190]{humanize_time(input_date)}[/]"


def Link(hyperlink: str, title: str) -> str:
    return f"[blue][link {hyperlink}]{title}[/]"


def Size(input_bytes: int) -> str:
    return f"[#3685B5]{human_size(input_bytes)}[/]"


@click.command()
@click.argument(
    "url", type=click.STRING, required=True, metavar="URL_OR_REPO_PATH"
)
@click.option(
    "--set-token", is_flag=True, default=False,
    help="Sets `url` to personal access token."
)
@click.option(
    "-l", "--long", is_flag=True, default=False,
    help="View more information."
)
@click.option(
    "-L", "--lang", is_flag=True, default=False,
    help="Show all languages of repo."
)
@click.option(
    "-f", "--file-tree", is_flag=True, default=False,
    help="Display files in a tree."
)
@click.option(
    "-p", "--path", type=click.STRING, default="", show_default=True,
    help="Set starting path for file tree relative to root (Github repo)."
)
@click.option(
    "-d", "--depth", type=click.IntRange(min=1), default=1, show_default=True,
    help="Depth to traverse file tree."
)
@click.option(
    "-b", "--branch", type=click.STRING, default="HEAD", show_default=True,
    help="Enter branch name or commit hash to view info or files from that specific branch/commit."
)
def main(url, **options):
    """
    Displays information on a Github repository.

    URL_OR_REPO_PATH must be either some form of Github Url or path
    starting with username and repo such as `user/repo/whatever`.
    """
    if options["set_token"]:
        set_token(url)
    elif options["file_tree"]:
        if options["depth"] == 1:
            depth = ""
        else:
            depth = f"... on Tree {{ entries {{ name type object {{... on Blob {{ byteSize }} }} }} }}"
            for i in range(options["depth"] - 2):
                depth = f"... on Tree {{ entries {{ name type object {{... on Blob {{ byteSize }} {depth} }} }} }}"

        token = get_token()
        owner, repo = get_url_info(url)
        data, rate_limit = run_query(
            FILE_QUERY_1 + depth + FILE_QUERY_2, token,
            {
                "owner": owner,
                "repo": repo,
                "path": f"{options['branch']}:{options['path']}"
            }
        )
        if list(data.keys())[0] == "errors":
            print(data["errors"][0]["message"])
            return
        try:
            entries = data["data"]["repository"]["object"]["entries"]
        except TypeError:
            print("Query failed. Make sure path is correct.")
            return

        entries = sort_entries(entries)
        root = populate_tree(f".{'/' if options['path'] else ''}{options['path']}", entries)

        for pre, fill, node in RenderTree(root, style=ContRoundStyle()):
            tree = "%s%s" % (pre, node.name)
            print(tree)

    elif options["lang"]:
        token = get_token()
        owner, repo = get_url_info(url)
        data, rate_limit = run_query(LANG_QUERY, token, {"owner": owner, "repo": repo, "branch": options["branch"]})
        if list(data.keys())[0] == "errors":
            print(data["errors"][0]["message"])
            return

        data = data["data"]["repository"]["languages"]

        total_size = data["totalSize"]
        total_count = data["totalCount"]

        grid = Table(
            show_header=False,
            header_style=None,
            box=ROUNDED_BORDER,
            title=f'[green]{owner}/{repo}[/] - Ratelimit: [blue]{rate_limit}[/]'
        )

        langs = []
        matrix = []
        for lang in data["edges"]:
            langs.append(
                f"[{lang['node']['color'] if lang['node']['color'] else '#FFFFFF'}]"
                f"{lang['node']['name']}[/] - "
                f"{Number(round(100 * lang['size'] / total_size, 2))}%"
            )

        start = 0
        end = 3

        for i in range(total_count // 3 + 1):
            matrix.append(langs[start:end])
            start += 3
            end += 3

        for row in matrix:
            grid.add_row(*row)

        print(grid)
    else:
        token = get_token()
        owner, repo = get_url_info(url)
        data, rate_limit = run_query(INFO_QUERY, token, {"owner": owner, "repo": repo, "branch": options["branch"]})
        if list(data.keys())[0] == "errors":
            print(data["errors"][0]["message"])
            return

        data = data["data"]["repository"]

        grid = Table(
            show_header=False,
            header_style=None,
            box=ROUNDED_BORDER,
            title=f'[green]{owner}/{repo}[/] - Ratelimit: [blue]{rate_limit}[/]'
        )

        latestRelease = data["latestRelease"]
        if not latestRelease:
            latestRelease = {
                "url": "",
                "name": "None"
            }

        licenseInfo = data["licenseInfo"]
        if not licenseInfo:
            licenseInfo = {
                "url": "",
                "spdxId": "None"
            }

        if options["long"]:
            grid.add_row(
                f"Owner          - {Link(f'https://github.com/{owner}', owner)} ",
                f"Created at     - {Date(data['createdAt'])} ",
                f"Is archived - {Bool(data['isArchived'])} "
            )
            grid.add_row(
                f"URL            - {Link(data['url'], 'Link')} ",
                f"Updated at     - {Date(data['updatedAt'])} ",
                f"Is disabled - {Bool(data['isDisabled'])} "
            )
            grid.add_row(
                f"License        - {Link(licenseInfo['url'], licenseInfo['spdxId'])} ",
                f"Pushed at      - {Date(data['pushedAt'])} ",
                f"Is fork     - {Bool(data['isFork'])} "
            )
            grid.add_row(
                f"Latest Release - {Link(latestRelease['url'], latestRelease['name'])} ",
                f"Disk usage     - {Size(data['diskUsage'] * 1024)} ",
                f"Is in org.  - {Bool(data['isInOrganization'])} "
            )
            grid.add_row(
                f"Forks          - {Number(data['forkCount'])} ",
                f"Watchers       - {Number(data['watchers']['totalCount'])} ",
                f"Is locked   - {Bool(data['isLocked'])} "
            )
            grid.add_row(
                f"Star count     - {Number(data['stargazerCount'])} ",
                f"Open Issues    - {Number(data['openIssues']['totalCount'])} ",
                f"Is mirror   - {Bool(data['isMirror'])} "
            )
            grid.add_row(
                f"Commit count   - {Number(data['object']['history']['totalCount'])} ",
                f"Closed Issues  - {Number(data['issues']['totalCount'] - data['openIssues']['totalCount'])}",
                f"Is private  - {Bool(data['isPrivate'])} "
            )
            grid.add_row(
                f"Open p.r.      - {Number(data['open_pr']['totalCount'])} ",
                f"Closed p.r.    - {Number(data['closed_pr']['totalCount'])} ",
                f"Merged p.r. - {Number(data['merged_pr']['totalCount'])} ",
            )
        else:
            grid.add_row(
                f"Owner    - {Link(f'https://github.com/{owner}', owner)} ",
                f"Disk usage - {Size(data['diskUsage'] * 1024)} ",
                f"Created at  - {Date(data['createdAt'])} ",
            )
            grid.add_row(
                f"URL      - {Link(data['url'], 'Link')} ",
                f"Stars      - {Number(data['stargazerCount'])} ",
                f"Updated at  - {Date(data['updatedAt'])} ",
            )
            grid.add_row(
                f"License  - {Link(licenseInfo['url'], licenseInfo['spdxId'])} ",
                f"Forks      - {Number(data['forkCount'])} ",
                f"Pushed at   - {Date(data['pushedAt'])} ",
            )
            grid.add_row(
                f"Language - [green]{data['languages']['nodes'][0]['name']}[/] ",
                f"Watchers   - {Number(data['watchers']['totalCount'])} ",
                f"Open issues - {Number(data['openIssues']['totalCount'])} "
            )
        print(grid)



from setuptools import setup, find_packages


def readme():
    with open("README.md") as f:
        _readme = f.read()
    return _readme


def requires():
    with open("requirements.txt") as f:
        _requires = f.read()
    return _requires


setup(
    name="gitinfo",
    version="1.0.3",
    description="Quickly get information about a Github repository",
    long_description=readme(),
    long_description_content_type="text/markdown",
    author="Folke Ishii",
    author_email="folke.ishii@gmail.com",
    license="MIT",
    python_requires=">=3.7",
    classifies=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Utilities"
    ],
    install_requires=find_packages('gitinfo'),
    packages=["gitinfo"],
    entry_points={
        "console_scripts": [
            "gitinfo = gitinfo.gitinfo:main",
        ]
    },
)
